import * as fs from 'fs';
import { promisify } from 'node:util';
import { walk as _walk } from '@nodelib/fs.walk';
import {} from 'fast-glob';
import fastGlob from 'fast-glob';
import picomatch from 'picomatch';
import { GLOBAL_IGNORE_PATTERNS } from '../constants.js';
import { debugLogObject } from './debug.js';
import * as path from './path.js';
import { timerify } from './Performance.js';
const walk = promisify(_walk);
function convertGitignoreToMicromatch(pattern, base) {
    let negated = pattern[0] === '!';
    if (negated) {
        pattern = pattern.slice(1);
    }
    const otherPatterns = [];
    if (!pattern.includes('/'))
        pattern = '**/' + pattern;
    else if (pattern.startsWith('/'))
        pattern = pattern.slice(1);
    if (pattern.endsWith('/'))
        otherPatterns.push(pattern + '**');
    else
        otherPatterns.push(pattern + '/**');
    return { negated, patterns: [pattern, ...otherPatterns].map(pattern => path.join(base, pattern)) };
}
function parseGitignoreFile(filePath, cwd) {
    const file = fs.readFileSync(filePath, 'utf8');
    const base = path.relative(cwd, path.dirname(path.toPosix(filePath)));
    return file
        .split(/\r?\n/)
        .filter(line => line && !line.startsWith('#'))
        .map(pattern => convertGitignoreToMicromatch(pattern, base));
}
async function _parseFindGitignores(options) {
    const ignores = ['.git', ...GLOBAL_IGNORE_PATTERNS];
    const unignores = [];
    const gitignoreFiles = [];
    const matcher = picomatch(ignores, { ignore: unignores });
    const entryFilter = (entry) => {
        if (entry.dirent.isFile() && entry.name === '.gitignore') {
            gitignoreFiles.push(entry.path);
            for (const rule of parseGitignoreFile(entry.path, options.cwd))
                if (rule.negated)
                    unignores.push(...rule.patterns);
                else
                    ignores.push(...rule.patterns);
            return true;
        }
        return false;
    };
    const deepFilter = (entry) => !matcher(path.relative(options.cwd, entry.path));
    await walk(options.cwd, {
        entryFilter: timerify(entryFilter),
        deepFilter: timerify(deepFilter),
    });
    debugLogObject(options.cwd, 'parsed gitignore files', { consideredFiles: gitignoreFiles, ignores, unignores });
    return { ignores, unignores };
}
const parseFindGitignores = timerify(_parseFindGitignores);
const cachedIgnores = new Map();
async function loadGitignores(options) {
    let gitignore = cachedIgnores.get(options.cwd);
    if (!gitignore) {
        gitignore = await parseFindGitignores(options);
        cachedIgnores.set(options.cwd, gitignore);
    }
    return gitignore;
}
export async function globby(patterns, options) {
    const ignore = options.gitignore && Array.isArray(options.ignore) ? [...options.ignore] : [];
    if (options.gitignore) {
        const gitignores = await loadGitignores(options);
        ignore.push(...gitignores.ignores);
        ignore.push(...gitignores.unignores.map(e => '!' + e));
    }
    debugLogObject(options.cwd, `fastGlobOptions`, { patterns, ...options, ignore });
    return fastGlob(patterns, {
        ...options,
        ignore,
    });
}
export async function isGitIgnoredFn(options) {
    cachedIgnores.clear();
    if (options.gitignore === false)
        return () => false;
    const gitignore = await loadGitignores(options);
    const matcher = picomatch(gitignore.ignores, { ignore: gitignore.unignores });
    const isGitIgnored = (filePath) => {
        const ret = matcher(path.relative(options.cwd, filePath));
        return ret;
    };
    return timerify(isGitIgnored);
}
